# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

"""
The method and partially also the code is reused from:
"The Intrinsic Dimensionaity of Images and Its Impact On Learning"
by Phillip Pope, Chen Zhu, Ahmed Abdelkader, Micah Goldblum, Tom Goldstein (ICLR 2021, spotlight)
code: https://github.com/ppope/dimensions
"""

import logging
import random
import warnings
from typing import Tuple

import numpy as np
import torch
from mml_dimensionality.models.knn import KNNComputer
from torch.utils.data import DataLoader
from tqdm import tqdm

from mml.core.data_loading.task_attributes import Modality
from mml.core.data_loading.task_dataset import TaskDataset

logger = logging.getLogger(__name__)


def calc_mle(cfg, dataset: TaskDataset) -> float:
    device = torch.device("cuda") if cfg.allow_gpu else torch.device("cpu")
    subsets = create_random_subsets(data_set=dataset, subset_size=cfg.sampling.sample_num)[: cfg.mode.max_subsets]
    if len(subsets[-1]) < cfg.mode.subset_min_size:
        subsets = subsets[:-1]
        if len(subsets) == 0:
            logger.error(
                f"Not enough samples to build a suitable subset. Current min size is "
                f"{cfg.mode.subset_min_size}, but task only offers {len(dataset)} samples!"
            )
            return float("Nan")
    subset_dims = []
    for s_set in tqdm(subsets, desc="Iterate subsets"):
        if len(s_set) < cfg.sampling.sample_num:
            warnings.warn(f"Subset only has {len(s_set)} samples!")
        model = run_knn(
            dataset=s_set, k=cfg.mode.k, batch_size=cfg.sampling.batch_size, n_workers=cfg.num_workers, device=device
        )
        dist = model.min_dists.cpu().numpy()

        mle_res, inv_mle_res = [], []
        for k in range(3, cfg.mode.k + 1):
            mle_results, invmle_results = intrinsic_dim_sample_wise_double_mle(k, dist)
            mle_res.append(mle_results.mean())
            inv_mle_res.append(1.0 / invmle_results.mean())
            logger.debug(f"Subset estimate for K={k}: MLE -> {mle_res[-1]}, INV-MLE -> {inv_mle_res[-1]}")
        if cfg.mode.inv_mle:
            subset_dims.append(inv_mle_res[-1])
        else:
            subset_dims.append(mle_res[-1])
    subset_dims = [d for d in subset_dims if np.isfinite(d)]
    if len(subset_dims) == 0:
        raise RuntimeError(
            "No valid subset dim found! Consider either increasing  mode.max_subsets, "
            "mode.subset_min_size or switching to mode.inv_mle=true"
        )
    logger.debug(f"all subset estimates: {subset_dims}")
    return sum(subset_dims) / len(subsets)


def run_knn(dataset: torch.utils.data.Subset, k: int, batch_size: int, n_workers: int, device: torch.device):
    model = KNNComputer(sample_num=len(dataset), k=k + 1)
    anchor_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=False, num_workers=n_workers)
    new_image_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=False, num_workers=n_workers)
    anchor_counter = 0
    model = model.to(device)
    # ignoring the labels
    with torch.no_grad():
        for a_batch in tqdm(anchor_loader, desc="Compute KNN", leave=False):
            a_images = a_batch[Modality.IMAGE.value].to(device)
            new_img_counter = 0
            for new_batch in new_image_loader:
                new_images = new_batch[Modality.IMAGE.value].to(device)
                # forward through model
                model(a_images, anchor_counter, new_images, new_img_counter)
                new_img_counter += new_images.size(0)

                # equiv_flag = (model.min_dists[anchor_start_idx:anchor_start_idx + a_images.size(0),
                #               0] == 0) & (model.min_dists[
                #                           anchor_start_idx:anchor_start_idx + a_images.size(0), 1] == 0)
                # if torch.any(equiv_flag):
                #     raise Exception("Identical data detected!")

            anchor_counter += a_images.size(0)
    return model.cpu()


def create_random_subsets(data_set, subset_size):
    indices = [i for i in range(len(data_set))]
    random.shuffle(indices)
    n_subsets = (len(data_set) // subset_size) + 1
    subset_idxes = [indices[i * subset_size : (i + 1) * subset_size] for i in range(n_subsets)]

    return [torch.utils.data.Subset(data_set, sidxes) for sidxes in subset_idxes]


def intrinsic_dim_sample_wise_double_mle(k: int, dist: np.ndarray) -> Tuple:
    """
    Returns Levina-Bickel dimensionality estimation and the correction by MacKay-Ghahramani.

    :param int k: nearest neighbors to use
    :param np.ndarray dist: array of the nearest distances
    :return: 2 dimensionality estimates
    """
    dist = dist[:, 1 : (k + 1)]
    assert np.all(dist > 0)
    d = np.log(dist[:, k - 1 : k] / dist[:, 0 : k - 1])
    d = d.sum(axis=1) / (k - 2)
    inv_mle = d.copy()
    d = 1.0 / d
    mle = d
    return mle, inv_mle



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from pathlib import Path

from hydra.core.config_search_path import ConfigSearchPath
from hydra.core.plugins import Plugins
from hydra.plugins.search_path_plugin import SearchPathPlugin

from mml.core.data_loading.file_manager import MMLFileManager

MMLFileManager.add_assignment_path(
    obj_cls=int,
    key="dimension",
    path=Path("PROJ_PATH") / "DIMENSIONS" / "TASK_NAME" / "dims.txt",
    enable_numbering=True,
    reusable=True,
)
MMLFileManager.add_assignment_path(
    obj_cls=None,
    key="dimension_plot",
    path=Path("PROJ_PATH") / "PLOTS" / "DIMENSION" / "dims_versus_perf.html",
    enable_numbering=True,
    reusable=False,
)


# register plugin configs
class MMLDimensionalitySearchPathPlugin(SearchPathPlugin):
    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        # Sets the search path for mml with copied config files
        search_path.append(provider="mml-dimensionality", path="pkg://mml_dimensionality.configs")


Plugins.instance().register(MMLDimensionalitySearchPathPlugin)



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from pathlib import Path
from typing import List

import pandas as pd
import plotly.express as px

from mml.core.data_loading.task_attributes import Keyword
from mml.core.data_loading.task_struct import TaskStruct
from mml.core.scripts.model_storage import ModelStorage


def plot_dimensions_versus_baseline(all_tasks: List[TaskStruct], store_path: Path) -> None:
    def load_dim(path: Path) -> int:
        with open(path, "r") as file:
            lines = file.readlines()
        assert len(lines) == 1
        return int(lines[0])

    def load_performance(models: List[ModelStorage]) -> float:
        metric = "val/Recall"
        return max([m.metrics[-1][metric] for m in models])

    domain_list = [
        Keyword.DERMATOSCOPY,
        Keyword.LARYNGOSCOPY,
        Keyword.GASTROSCOPY_COLONOSCOPY,
        Keyword.LAPAROSCOPY,
        Keyword.NATURAL_OBJECTS,
        Keyword.HANDWRITINGS,
        Keyword.CATARACT_SURGERY,
        Keyword.FUNDUS_PHOTOGRAPHY,
        Keyword.MRI_SCAN,
        Keyword.X_RAY,
        Keyword.CT_SCAN,
        Keyword.CLE,
        Keyword.CAPSULE_ENDOSCOPY,
        Keyword.ULTRASOUND,
    ]

    def get_domain(task: TaskStruct) -> Keyword:
        domain_candidates = [d for d in task.keywords if d in domain_list]
        # TODO this is a hacky solution only so far
        if task.name == "svhn":
            domain_candidates.append(Keyword.NATURAL_OBJECTS)
        if task.name == "mnist_digit_classification" or task.name == "emnist_digit_classification":
            domain_candidates.append(Keyword.HANDWRITINGS)
        if len(domain_candidates) != 1:
            raise RuntimeError(f"{task} with {domain_candidates}")
        return domain_candidates[0]

    task_infos = [
        {
            "name": task.name,
            "dimension": load_dim(task.paths["dimension"]),
            "performance": load_performance(task.models),
            "domain": get_domain(task).value,
        }
        for task in all_tasks
    ]
    df = pd.DataFrame(data=task_infos)
    fig = px.scatter(data_frame=df, x="dimension", y="performance", color="domain", hover_name="name", trendline="ols")
    fig.write_html(store_path)



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

# automatically activate plugin when importing
import mml_dimensionality.activate  # noqa



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

"""
The method and partially also the code is reused from:
"The Intrinsic Dimensionaity of Images and Its Impact On Learning"
by Phillip Pope, Chen Zhu, Ahmed Abdelkader, Micah Goldblum, Tom Goldstein (ICLR 2021, spotlight)
code: https://github.com/ppope/dimensions
"""

import torch


class KNNComputer(torch.nn.Module):
    """
    Using this hack for data parallel
    without checking for the sample itself
    """

    def __init__(self, sample_num: int, k: int = 1, cosine_dist=False):
        super(KNNComputer, self).__init__()

        self.K = k
        self.cosine_dist = cosine_dist
        self.register_buffer("num_computed", torch.zeros([]))

        if k == 1:
            self.register_buffer("min_dists", torch.full((sample_num,), float("inf")))
            self.register_buffer("nn_indices", torch.full((sample_num,), 0, dtype=torch.int64))
        else:
            self.register_buffer("min_dists", torch.full((sample_num, k), float("inf")))
            self.register_buffer("nn_indices", torch.full((sample_num, k), 0, dtype=torch.int64))

    def forward(self, x, x_idx_start, y, y_idx_start):
        # update the min dist for existing examples...
        x_bsize, y_bsize = x.size(0), y.size(0)
        x = x.view(x_bsize, -1)
        y = y.view(y_bsize, -1)
        if self.cosine_dist:
            x = x / x.norm(dim=1, keepdim=True)
            y = y / y.norm(dim=1, keepdim=True)
            dist = x.mm(y.t())

        else:
            # dist = torch.norm(x.unsqueeze(1) - y.unsqueeze(0), dim=2)
            dist = torch.cdist(x, y, p=2, compute_mode="donot_use_mm_for_euclid_dist")

        if self.K == 1:
            new_min_dist, nn_idxes = torch.min(dist, dim=1)

            self.min_dists[x_idx_start : x_idx_start + x_bsize] = torch.min(
                new_min_dist, self.min_dists[x_idx_start : x_idx_start + x_bsize]
            )
            self.nn_indices[x_idx_start : x_idx_start + x_bsize] = nn_idxes + y_idx_start
        else:
            comp = torch.cat([dist, self.min_dists[x_idx_start : x_idx_start + x_bsize]], dim=1)
            # updated_min_dist, nn_idxes = torch.topk(comp, self.K, dim=1, largest=False)
            # check for repeated images
            sorted_dists, sorted_idxes = torch.sort(comp, dim=1, descending=False)
            updated_dist_list, nn_idx_list = [], []
            for row in range(sorted_dists.shape[0]):
                sidx = 1
                while sidx < sorted_dists.shape[1]:
                    if sorted_dists[row, sidx] == 0:
                        sidx += 1
                    else:
                        break
                updated_dist_list.append(sorted_dists[row, sidx - 1 : sidx - 1 + self.K])
                nn_idx_list.append(sorted_idxes[row, sidx - 1 : sidx - 1 + self.K])
            updated_min_dist = torch.stack(updated_dist_list)
            nn_idxes = torch.stack(nn_idx_list)

            self.min_dists[x_idx_start : x_idx_start + x_bsize] = updated_min_dist

            sample_idxes = (nn_idxes < y_bsize).int() * (nn_idxes + y_idx_start) + (
                nn_idxes >= y_bsize
            ).int() * self.nn_indices[x_idx_start : x_idx_start + x_bsize]
            self.nn_indices[x_idx_start : x_idx_start + x_bsize] = sample_idxes

    def get_mean_nn_dist(self, sidx, eidx):
        if self.K == 1:
            return torch.mean(self.min_dists[sidx:eidx])



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

from pathlib import Path


def load_dim(path: Path) -> int:
    """
    Loads the dimensionality of a task, provided the storage path of that information.

    :param Path path:
    :return:
    """
    with open(path, "r") as file:
        lines = file.readlines()
    assert len(lines) == 1
    return int(lines[0])



# LICENSE HEADER MANAGED BY add-license-header
#
# SPDX-FileCopyrightText: Copyright 2024 German Cancer Research Center (DKFZ) and contributors.
# SPDX-License-Identifier: MIT
#

import logging

import numpy as np
from mml_dimensionality.scripts.mle import calc_mle
from mml_dimensionality.visualization import plot_dimensions_versus_baseline
from omegaconf import DictConfig

from mml.core.scripts.decorators import beta
from mml.core.scripts.exceptions import MMLMisconfigurationException
from mml.core.scripts.schedulers.base_scheduler import AbstractBaseScheduler
from mml.core.scripts.utils import LearningPhase

logger = logging.getLogger(__name__)


@beta("Dimension estimation is in beta!")
class DimensionalityScheduler(AbstractBaseScheduler):
    """ "
    AbstractBaseScheduler implementation for the estimation of task dimensionality. Includes the following subroutines:
    - estimate

    The method and partially also the code is reused from:
    "The Intrinsic Dimensionaity of Images and Its Impact On Learning"
    by Phillip Pope, Chen Zhu, Ahmed Abdelkader, Micah Goldblum, Tom Goldstein (ICLR 2021, spotlight)
    code: https://github.com/ppope/dimensions
    """

    def __init__(self, cfg: DictConfig):
        # initialize
        super(DimensionalityScheduler, self).__init__(cfg=cfg, available_subroutines=["estimate", "plot"])
        if self.cfg.mode.k < 3:
            raise MMLMisconfigurationException("mode.k must be at least 3.")
        if self.cfg.augmentations.normalization is not None:
            raise MMLMisconfigurationException(
                "Must deactivate normalization for dimensionality estimation. " "Best set augmentations=no_norm."
            )

    def create_routine(self):
        """
        This scheduler implements one subroutine, which estimates a task's dimensionality.

        :return: None
        """
        # -- add preprocess command
        if "estimate" in self.subroutines:
            if self.pivot:
                logger.info("Dimensionality mode with pivot task will only estimate this task!")
                self.commands.append(self.estimate_task_dimensionality)
                self.params.append([self.pivot])
            else:
                for task in self.cfg.task_list:
                    self.commands.append(self.estimate_task_dimensionality)
                    self.params.append([task])

    def before_finishing_hook(self):
        if "plot" in self.subroutines and not self.pivot:
            task_list = []
            # check if performances are available
            for task in self.cfg.task_list:
                struct = self.get_struct(task_name=task)
                if len(struct.models) > 0:
                    task_list.append(struct)
                else:
                    logger.error(
                        f"No performance found for task {task} to produce dimension versus performance plot, "
                        f"you may want to use reuse.models=SOME_PROJ_NAME to provide performances."
                    )

            if len(task_list) > 0:
                plot_path = self.fm.construct_saving_path(obj=None, key="dimension_plot")
                plot_dimensions_versus_baseline(all_tasks=task_list, store_path=plot_path)
                logger.info(f"Plotted task dimensionality against performance at {plot_path}.")

    def estimate_task_dimensionality(self, task_name: str):
        logger.info("Starting estimating dimensionality data for task " + self.highlight_text(task_name))
        task_struct = self.get_struct(task_name)
        datamodule = self.create_datamodule(task_structs=task_struct)
        datamodule.setup(stage="fit")
        dataset = datamodule.task_datasets[task_name][LearningPhase.TRAIN]
        val = calc_mle(cfg=self.cfg, dataset=dataset)
        dim = -1 if np.isnan(val) else int(val)
        path = self.fm.construct_saving_path(obj=dim, key="dimension", task_name=task_struct.name)
        with open(path, "w+") as f:
            f.write(str(dim))
        task_struct.paths["dimension"] = path
        self.task_factory.dump()
        logger.info("Finished dimensionality estimation for task " + self.highlight_text(task_name))
